from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from past.builtins import basestring

from datetime import datetime
import logging
from urllib.parse import urlparse
from time import sleep
import re
import sys
import subprocess
import pdb

import airflow
from airflow import hooks, settings
from airflow.exceptions import AirflowException, AirflowSensorTimeout, AirflowSkipException
from airflow.models import BaseOperator, TaskInstance
from airflow.hooks.base_hook import BaseHook
from airflow.hooks.hdfs_hook import HDFSHook
from airflow.utils.state import State
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults


class gliteSensor(BaseSensorOperator):
    """
    Runs a sql statement until a criteria is met. It will keep trying until
    sql returns no row, or if the first cell in (0, '0', '').

    :param conn_id: The connection to run the sensor against
    :type conn_id: string
    :param sql: The sql to run. To pass, it needs to return at least one cell
        that contains a non-zero / empty string value.
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#7c7287'

    @apply_defaults
    def __init__(self, 
            submit_task, 
            success_threshold=0.9, 
            poke_interval=120,
            timeout=60*60*24*4, 
            *args, **kwargs):
        self.submit_task= submit_task
        self.threshold=success_threshold
        self.glite_status='Waiting'
        super(gliteSensor, self).__init__(poke_interval=poke_interval,
                timeout=timeout, *args, **kwargs)

    def poke(self, context):
        self.jobID=context['task_instance'].xcom_pull(task_ids=self.submit_task)
        if self.jobID==None:
            raise RuntimeError("Could not get the jobID from the "+str(self.submit_task)+" task. ")
        logging.info('Poking glite job: ' + self.jobID)
        g_proc = subprocess.Popen(['glite-wms-job-status', self.jobID] ,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        g_result=g_proc.communicate() 
        self.parse_glite_jobs(g_result[0])
        if not 'Done' in self.job_status:
            if 'Abort' in self.job_status:
                logging.warn("Job aborted from commandline")
                return True 
            return False
        else:
            exit_codes=self.count_successes(g_result[0])
            success_rate=1
            logging.info(str(success_rate)+" of jobs completed ok")
            if (success_rate < self.threshold):
                logging.warn("Less than "+str(self.threshold)+" jobs finished ok!")
            return True

    def parse_glite_jobs(self,jobs): 
        try:
            self.job_status=jobs.split('Current Status:')[1].split()[0]
        except:
            logging.info(jobs)
        logging.debug("Current job status is "+str(self.job_status))
        if self.glite_status== 'Running': 
            self.ui_color='#ef7f23'
        if self.glite_status=='Waiting':
            self.count_successes(jobs)
        if self.glite_status=='Running' and self.job_status=='Waiting':
            self.glite_status='Completed'


    def count_successes(self,jobs):
        """Counts the number of Completed jobs in the results of the glite-wms-job-status
        output.  """
        exit_codes=[]
        jobs_list=[]
        for j in jobs.split('=========================================================================='):
            jobs_list.append(j)
        statuses=[]
        for j in jobs_list:
            if "Current Status:" in j:
                statuses.append(j.split("Current Status:")[1].split('\n')[0])
        numdone=0
        for i in statuses:
            if 'Done' in i or 'Cancelled' in i or 'Aborted' in i  :
                numdone+=1
        if 'Done' in statuses[0]: 
            self.job_status = 'Done'
        if numdone == len(jobs):
            self.job_status='Done'
        if self.job_status == 'Waiting':
            for i in statuses:
                if 'Scheduled' in i or 'Running' in i:
                    self.job_status = 'Waiting'
                    return  statuses[1:]
                self.job_status = "Done"
        logging.info("Num_jobs_done "+str(numdone))
        return statuses[1:]
        



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from builtins import bytes
import os
import os.path
import signal
from time import sleep
import logging
from subprocess import Popen, STDOUT, PIPE
from tempfile import gettempdir, NamedTemporaryFile


from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.utils.file import TemporaryDirectory
from airflow.utils.state import State
#import progressbar #TODO: Either add progressbar to install or ignore this
#logging.info(progressbar.__file__)
import GRID_LRT.Application.submit as submit
import GRID_LRT.Token as Token
from GRID_LRT import get_picas_credentials
import pdb

class LRTSubmit(BaseOperator):
    """
    Execute a stage command using the LOFAR API. Input is either file with srms
        a list of srms or both.

    :param srmfile: the name of the file holding a list of srm files to stage
    :type srmfile: string
    :param srms: a list of the srms that need to be staged
    :type srms: list
    :param stageID: In case staging was already done
    :type stageID: string
    :type output_encoding: output encoding of bash command
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#f0ede4'

    @apply_defaults
    def __init__(
            self,
            token_task,
            NCPU=2,
            parameter_step=4,
            run_location='sara',
            output_encoding='utf-8',
            *args, **kwargs):

        super(LRTSubmit, self).__init__(*args, **kwargs)
        self.token_task = token_task 
        self.location = run_location
        self.parameter_step = parameter_step
        self.NCPU = NCPU
        self.output_encoding = output_encoding
        self.state = State.QUEUED

    def execute(self, context):
        """
        Execute the bash command in a temporary directory
        which will be cleaned afterwards
        """
        pc=get_picas_credentials.picas_cred()
        self.token_id=context['task_instance'].xcom_pull(task_ids=self.token_task)['token_type']
        
        if self.token_id==None:
            raise RuntimeError("Could not get the token list from the "+str(self.token_task)+ " task")

        self.initialilze_submitter(location = self.location,
                NCPU = self.NCPU, parameter_step = self.parameter_step)
        pc=get_picas_credentials.picas_cred()
        th=Token.Token_Handler(t_type=self.token_id,
                    uname=pc.user,pwd=pc.password,dbn=pc.database)
        token_list=th.list_tokens_from_view('todo')
        
        self.submitter.numjobs=int(len(token_list)) 
        if len(token_list)<1:
            raise("Not enough todo tokens to submit!")
        with self.submitter as sub:
            launch_ID=sub.launch()
        return launch_ID

    def success(self):
        self.status=State.SUCCESS
        logging.info("Successfully staged " +
                    str(self.progress['Percent done']) + " % of the files.")

    def initialilze_submitter(self,location='sara',**kwargs):
        self.submitter=None 
        if location=='sara':
            self.submitter=submit.jdl_launcher(token_type=self.token_id,**kwargs)


    def get_token_list(self,context):
        token_id=context['task_instance'].xcom_pull(task_ids=self.token_task)['token_type']


    def on_kill(self):
        logging.warn('Sending SIGTERM signal to staging group')
        self.state=State.SHUTDOWN
        os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)



# -*- coding: utf-8 -*-                                                                                  
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from functools import wraps

import math

import logging                                                                                           
import subprocess
import os
import re
from os import fdopen, remove
from shutil import move

import pdb

from airflow.utils.decorators import apply_defaults                                                      
from airflow.models import Variable
from airflow import settings
from airflow.utils.log.logging_mixin import LoggingMixin
from GRID_LRT.Staging.srmlist import srmlist
import tempfile
from tempfile import mkstemp
log = LoggingMixin().log

Field_Statuses={'AAA':'Starting Processing',
                'SSS':'Staging done',
                'CCC':'Calibrator Done',
                'TTT':'Target Done',
                'RRR':'Error',
                'ZZZ':'Completed'}

def get_user_proxy(username):
    """
    Gets the X509 file by the user in order to perform authorized operations by them
    as opposed as by the DAG Executor
    """
    pass

def make_srmfile_from_step_results(prev_step_token_task):
    pass

def check_if_enoug_output_files(outp_task):
    pass

def modify_parset(parset_path, freq_res, time_res, OBSID, flags ):
    """Takes in a base_parset path and changes the time and frequency resolution parameters 
    of this parset. Saves it into a tempfile. Returns the tempfile_path"""
    fh, abs_path = mkstemp()
    with open(parset_path, 'r') as file:
        filedata = file.read()
        filedata = re.sub(r'\! avg_timestep\s+=\s\S?',
                                      "! avg_timestep         = "+str(int(time_res)), filedata)
        filedata = re.sub(r'\! avg_freqstep\s+=\s\S?',
                                              "! avg_freqstep         = "+str(int(freq_res)), filedata)
        filedata = re.sub(r'\! flag_baselines\s+=\s+\W\s+\S+\s+\W',
                                              "! flag_baselines    = "+flags+ "\n", filedata)
        file_ending=".MS"
        if 'gsmcal_solve' in filedata:
            file_ending='.ms'
        filedata = re.sub(r'\! target_input_pattern\s+=\s\S+',
                                  "! target_input_pattern    = " + str(OBSID) + "*" + file_ending,  filedata)
        if 'calib_cal' in filedata:
            file_ending=".MS"
        if 'h5imp_cal' in filedata:
            file_ending=".ndppp_prep_cal"
        filedata = re.sub(r'\! cal_input_pattern\s+=\s\S+',
                              "! cal_input_pattern    = "+str(OBSID)+"*"+file_ending,
                            filedata)

    with open(abs_path, 'w') as pars_file:
        pars_file.write(filedata)
        os.chmod(abs_path, 0o666 )
    return(abs_path)


def modify_parset_from_fields_task(parsets_dict={}, fields_task=None, time_avg=8, freq_avg=2, flags=None,  **context):
    """Takes a dict of 'original' parsets and the task with the fields information
    which will be used to update the averaging paremetes. Returns a dict of the 'name':'location' of the 
    modified parsets, so they can be used by the upload tokens task"""
    field_dict = context['ti'].xcom_pull(fields_task)
    targ_freq = int(field_dict['targ_freq_resolution'])/freq_avg
    targ_time = time_avg/int(math.floor(float(field_dict['targ_time_resolution'])))
    cal_freq = int(field_dict['calib_freq_resolution'])/freq_avg
    cal_time = time_avg/int(math.floor(float(field_dict['calib_time_resolution'])))
    cal_OBSID = field_dict['calib_OBSID']
    targ_OBSID = field_dict['target_OBSID']
    resulting_parsets = {}
    if not flags:
        flags = ", ".join(field_dict['baseline_filter'].split(" ")) 
    for p_name, p_path in parsets_dict.items():
        if "Calib" in p_name:
            resulting_parsets[p_name] = modify_parset(p_path, cal_freq, cal_time, cal_OBSID, flags)
        if "Targ" in p_name:
            resulting_parsets[p_name] = modify_parset(p_path, targ_freq, targ_time, targ_OBSID, flags)
    return resulting_parsets


def set_user_proxy_var(proxy_location):
    os.environ['X509_USER_PROXY']=proxy_location


def set_field_status(fields_file, cal_OBSID, targ_OBSID, field_name, status):
    fh, abs_path = mkstemp()
    with fdopen(fh,'w') as tmp_f:
        with open(fields_file,'r') as f:
            for line in f:
                if line.split(',')[11] == field_name and "L"+line.split(',')[1] == targ_OBSID and "L"+line.split(',')[6] == cal_OBSID:
                    tmp_f.write(str(','.join([str(status)]+line.split(',')[1:])))
                else:
                    tmp_f.write(line)
    remove(fields_file)
    move(abs_path, fields_file)
    os.chmod(fields_file, 0o666 )


def set_field_status_from_taskid(fields_file, task_id, status, **context):
    """sets the field status as the status input variable"""
    field_data=context['ti'].xcom_pull(task_id)
    field_name = field_data['field_name']
    cal_OBSID = field_data['calib_OBSID']
    targ_OBSID = field_data['target_OBSID']
    set_field_status(fields_file, cal_OBSID, targ_OBSID, field_name, status)


def set_field_status_from_task_return(fields_file, task_id, status_task, **context):
    """Sets the field status based on the (String) reuturned by the status_task variable

    """
    field_data=context['ti'].xcom_pull(task_id)
    return_data=context['ti'].xcom_pull(status_task)
    field_name = field_data['field_name']
    cal_OBSID = field_data['calib_OBSID']
    targ_OBSID = field_data['target_OBSID']
    set_field_status(fields_file, cal_OBSID, targ_OBSID, field_name, return_data)


def get_field_location_from_srmlist(srmlist_task, srmfile_key='targ_srmfile', **context):
    """Gets the srmlist from a task and returns the location of the field
    IE the LTA location where the raw data is stored"""
    field_loc="U" #U=unknown
    srm_data=context['ti'].xcom_pull(srmlist_task)
    srmfile=srm_data[srmfile_key]
    _s_list=srmlist()
    for i in  open(srmfile,'r').read().split():
        _s_list.append(i)
    if _s_list.LTA_location == 'juelich':
        return "JD"
    if _s_list.LTA_location == 'sara':
        return "SD"
    if _s_list.LTA_location == 'poznan':
        return "PD"
    return "UNK"


def check_if_running_field(fields_file):
    """
    Checks if there are fields that are running (ie not completed, or Error)
    Returns True if running, False if not
    """
    for i in open(fields_file,'r').read().split():
        status=i.split(',')[0]
        if status=='AAA' or status=='SSS' or status=='CCC':
            return True
    return False

def get_srmfile_from_dir(srmdir,field_task, **context):
    field_data=context['ti'].xcom_pull(field_task)
    cal_OBSID=field_data['calib_OBSID']
    targ_OBSID=field_data['target_OBSID']
    calib_srmfile=srmdir+'srm'+str(cal_OBSID.split('L')[-1])+".txt"
    targ_srmfile=srmdir+'srm'+str(targ_OBSID.split('L')[-1])+".txt"
    Variable.set("SKSP_Prod_Calibrator_srm_file",calib_srmfile)
    Variable.set("SKSP_Prod_Target_srm_file",targ_srmfile)
    logging.info("Calib_srmfile is "+calib_srmfile)
    logging.info("Targ_srmfile is "+targ_srmfile)
    return {'cal_srmfile':calib_srmfile,
            'targ_srmfile':targ_srmfile, 
            "CAL_OBSID": cal_OBSID, 
            "TARG_OBSID": targ_OBSID}


def get_next_field(fields_file, indicator='SND', **context):
    """Determines the next field to be processed,
    uses the set_field_status function to set its status to
    started().
    By default it locks SARA tokens (marked SND in the first column of the
    fields_file), however using the indicator variable, other files can be locked

    """
    for i in open(fields_file,'r').read().split('\n'):
        if not i.split(',')[0] == indicator:
            continue
        l=i.split(',')
        break
    field_information = {'target_OBSID':'L'+l[1],
            'targ_freq_resolution':l[3],
            'targ_time_resolution':l[4],
            'calib_OBSID':'L'+l[6],
            'calib_freq_resolution':l[8],
            'calib_time_resolution':l[9],
            'field_name':l[11],
            'baseline_filter':l[12]}
 
    logging.info("Field Information is: "+str(field_information))
    return field_information


def count_files_uberftp(directory):
    num_files=0
    logging.info(directory)
    c=subprocess.Popen(['uberftp','-ls', directory],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out=c.communicate()
    if out[1]!='':
        logging.warning("srmls failed to find a directory!! ")
        return 0
    if 'lofsksp' not in out[0]:
        logging.warning("no link found in folder!, returning 0 files")
        return 0
    file_list=[directory+"/"+str(i.split()[-1])
                            for i in out[0].strip().split("\r\n")]
    return len(file_list)

def get_var_from_task_decorator(Cls, upstream_task_id="", upstream_return_key="",u_task=None):
    """wrapper for functions that require an fixed input, which is here provided
    by a previous task"""
#    upstream_return=context['ti'].xcom_pull(task_ids=upstream_task_id)
    upstream_return={upstream_return_key:"test"}
    return_value=upstream_return[upstream_return_key]
    Cls.srmfile=return_value
    return Cls


def count_from_task(srmlist_task, srmfile_name, task_if_less,
                    task_if_more, pipeline="SKSP",step='pref_cal2',
                    min_num_files=1, **context):
    srmlist=context['ti'].xcom_pull(task_ids=srmlist_task)[srmfile_name]
    return(count_grid_files(srmlist, task_if_less,task_if_more, pipeline=pipeline,step=step, min_num_files=min_num_files))


def check_folder_for_files(folder,number=1):
    """Raises an exception (IE FAILS) if the (gsiftp) folder has less than
    'number' files
    
    By default fails if folder is empty"""
    num_files = count_files_uberftp(folder)
    if num_files < number:
        raise ValueError("Not enough Files ("+
                str(num_files)+"<"+str(number)+") in "+ str(folder))
    return

def check_folder_for_files_from_task(taskid, xcom_key, number, **context):
    path = context['ti'].xcom_pull(task_ids=taskid)[xcom_key]
    check_folder_for_files(path,number)
    

def count_grid_files(srmlist_file, task_if_less,
        task_if_more, pipeline="SKSP",step='pref_cal1',min_num_files=1):
    """
    An airflow function that branches depending on whether there are calibrator or target
    solutions matching the files in the srmlist. 
    """ 
    srml=srmlist()
    for line in open(srmlist_file).readlines():
        srml.append(line)
    num_files = count_files_uberftp('gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines/'+str(pipeline)+'/'+str(step)+'/'+str(srml.OBSID))
    if num_files >= min_num_files:
        return task_if_more
    else:
        return task_if_less


def stage_if_needed(stage_task, run_if_staged, run_if_not_staged,
                     **context):
    """This function takes the check_if_staged task and stages the files
    if None is returned. Otherwise it passes the srmlist to the 'join' task
    and the processing continues
    """
    staged_status=context['ti'].xcom_pull(task_ids=stage_task)
    logging.info(staged_status)
    if staged_status['staged']:
        return run_if_staged
    else:
        return run_if_not_staged

def delete_gsiftp_files(gsiftp_directory):
    del_task = subprocess.Popen(['uberftp','rm',gsiftp_directory+"/*"], stdout=subprocess.PIPE, 
            stderr=subprocess.PIPE)
    output = del_task.communicate()
    if output[1] != '':
        logging.warn("Error deleting files:")
        logging.warn(output[1])
    return gsiftp_directory

def delete_gsiftp_from_task(root_dir, OBSID_task, **context):
    """A task that returns OBSID, and a root directory come together in this
    function to gracefully delete all files in this here directory

    """
    task_results=context['ti'].xcom_pull(task_ids=OBSID_task)
    OBSID=task_result['OBSID']
    logging.debug('Deleting files from directory '+str(root_dir)+"/"+str(OBSID))
    delete_gsiftp_files(root_dir+"/"+OBSID_task)



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from past.builtins import basestring

from datetime import datetime
import logging
from urllib.parse import urlparse
from time import sleep
import re
import os
import sys
import subprocess
import pdb
from tempfile import gettempdir, NamedTemporaryFile

from GRID_LRT import Token
from GRID_LRT import get_picas_credentials

import airflow
from airflow import hooks, settings
from airflow.exceptions import AirflowException, AirflowSensorTimeout, AirflowSkipException
from airflow.models import BaseOperator, TaskInstance
from airflow.hooks.base_hook import BaseHook
from airflow.hooks.hdfs_hook import HDFSHook
from airflow.utils.state import State
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults


class Storage_to_Srmlist(BaseOperator):
    """
    TODO: srmls -l on the base directory and parse locality instead of looping!


    :param conn_id: The connection to run the sensor against
    :type conn_id: string
    :param sql: The sql to run. To pass, it needs to return at least one cell
        that contains a non-zero / empty string value.
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#7c7287'

    @apply_defaults
    def __init__(self, 
            srm_directory=None,
            token_task=None,
            srmfile=None, 
            append=False,
            *args, **kwargs):
        self.srm_dir=None
        if srm_directory:
            self.srm_dir=srm_directory
        else:
            if not token_task:
                raise ValueError("srm_directory not defined") 
            else:
                self.token_task=token_task
        super(Storage_to_Srmlist, self).__init__(*args, **kwargs)
        if srmfile:
            self.srmfile=srmfile #Blank out file
            if not append:
                with open(self.srmfile,'w') as f:
                    f.write()
        else:
            self.tempfile=NamedTemporaryFile(delete=False)
            self.srmfile=self.tempfile.name  

    def execute(self, context):
        if not self.srm_dir:
            token_task_result=context['task_instance'].xcom_pull(task_ids=self.token_task)
            self.srm_dir=self.get_srmdir_from_token_task(token_task_result)
        logging.info("getting files from directory: "+str(self.srm_dir))
        srm_links=self.list_dir(self.srm_dir)
        with open (self.srmfile,'wb') as writefile:
            for i in srm_links:
                writefile.write(str(i)+ '\n')
        logging.info("gsiftp srms stored in "+str(self.srmfile))
        logging.info("last link is "+str(i))
        return {'srmfile':self.srmfile, 'staged':True}

    def get_srmdir_from_token_task(self, token_task_view):
        OBSID=token_task_view['OBSID']
        t_type=token_task_view['token_type']
        view=token_task_view['view']
        key='RESULTS_DIR'
        pc=get_picas_credentials.picas_cred()
        th=Token.Token_Handler(t_type=t_type, uname=pc.user, pwd=pc.password, dbn=pc.database)
        tokens=th.list_tokens_from_view(view)
        srmdir=''
        for t in tokens:
            if srmdir!='': break
            srmdir=str(th.db[t['id']][key])+str(OBSID) 
        return srmdir

    def list_dir(self,srmdir):
        """Checks the status of all srm links in the folder given. 
        Returns a list of localities
        """
        g_proc = subprocess.Popen(['uberftp','-ls', srmdir] ,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out=g_proc.communicate()
        if out[1]!='':
            raise RuntimeError('srmls failed') 
        if 'lofsksp' not in out[0]: 
            logging.warning("no link found in folder!")
        file_list=[srmdir+"/"+str(i.split()[-1]) 
                for i in out[0].strip().split("\r\n")]
        return file_list



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from builtins import bytes
import os
import os.path
import signal
from time import sleep
import logging
from subprocess import Popen, STDOUT, PIPE
from tempfile import gettempdir, NamedTemporaryFile, mkdtemp
import tarfile
import shutil

from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.utils.file import TemporaryDirectory
from airflow.utils.state import State
#import progressbar
from GRID_LRT import Token
from GRID_LRT import get_picas_credentials
from GRID_LRT.Staging.srmlist import srmlist
from GRID_LRT.Staging.srmlist import slice_dicts
import yaml

import pdb

class TokenCreator(BaseOperator):
    """
    Using a Token template input, this class creates the tokens for a LOFAR job
    The tokens are a set of documents that contain the metadata for each processing
    job as well as the job's progress, step completion times, and etc. 

    :param sbx_task: The name of the sandbox task which passes the sbx_loc to the tokens
    :type sbx_task: string
    :param srms: a list of the srms that need to be staged
    :type srms: list
    :param stageID: In case staging was already done
    :type stageID: string
    :type output_encoding: output encoding of bash command
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#f3f92c'

    @apply_defaults
    def __init__(
            self,
            tok_config,
            sbx_task,
            staging_task,
            srms_task = None,
            fields_task = None,
            pc_database=None,
            subband_prefix = 'SB',
            subband_suffix = '_',
            token_type = 'test_',
            files_per_token = 10,
            output_encoding = 'utf-8',
            *args, **kwargs):

        super(TokenCreator, self).__init__(*args, **kwargs)
        self.pc_database = pc_database
        self.tok_config  =  tok_config  
        self.sbx_task = sbx_task
        self.fields_task = fields_task
        self.subband_prefix = subband_prefix
        self.subband_suffix = subband_suffix
        self.staging_task = staging_task
        self.srms_task = srms_task
        self.files_per_token = files_per_token
        self.output_encoding  =  output_encoding
        self.t_type = token_type
        self.state = State.QUEUED

    def execute(self, context):
        """
        Execute the bash command in a temporary directory
        which will be cleaned afterwards
        """
        srms = self.get_staged_srms(context)
        if not srms:
            print("Could not get the list of staged srms!")
        pc = get_picas_credentials.picas_cred()
        if self.pc_database:
            pc.database = self.pc_database
        if self.fields_task:
            app = context['task_instance'].xcom_pull(task_ids=self.fields_task)['field_name']
        else:
            app = srms.OBSID
        self.t_type= self.t_type+app
        tok_settings = yaml.load(open(self.tok_config,'rb'))['Token']
        pipe_type = tok_settings['PIPELINE_STEP']
        th = Token.Token_Handler(t_type=self.t_type,
                    uname=pc.user,pwd=pc.password,dbn=pc.database)
        th.add_overview_view()
        th.add_status_views()
        th.add_view(view_name = pipe_type, cond='doc.PIPELINE_STEP == "{0}" '.format(pipe_type), emit_value2='doc.status')
        
        logging.info('Token type is '+self.t_type)
        logging.info('Tokens are available at https://picas-lofar.grid.surfsara.nl:6984/_utils/database.html?'+pc.database+'/_design/'+self.t_type+'/_view/overview_total')
        logging.info("Token settings are :")
        for i in tok_settings.items():
            logging.info(str(i))

        self.tokens = Token.TokenSet(th=th,tok_config=self.tok_config)
        self.upload_tokens(self.tokens)
        logging.debug(srms)
        if self.files_per_token != 1:
            d = slice_dicts(srms.sbn_dict(pref=self.subband_prefix, 
                                    suff=self.subband_suffix)
                                    ,self.files_per_token)
        else:
            d = {}
            for i in srms.sbn_dict(pref=self.subband_prefix, suff=self.subband_suffix):
                d[i[0]] = i[1]
        self.tokens.create_dict_tokens(iterable=d,
                id_append=pipe_type,key_name='STARTSB',
                file_upload='srm.txt')
        self.tokens.add_keys_to_list("OBSID",srms.OBSID)
        
        if 'CAL_OBSID' in tok_settings.keys():
            cal_results = context['task_instance'].xcom_pull(task_ids=self.srms_task) 
            CAL_OBSID = cal_results['CAL_OBSID']
            self.tokens.add_keys_to_list("CAL_OBSID",CAL_OBSID)
        
        sbx_name = context['task_instance'].xcom_pull(task_ids=self.sbx_task)["SBX_location"]
        self.tokens.add_keys_to_list('SBXloc',"gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/sandbox/"+sbx_name)
        results = dict()
        results['output_dir'] = tok_settings['RESULTS_DIR']+"/"+ str(srms.OBSID)
        results['token_type'] = str(self.t_type)
        results['view'] = pipe_type
        results['OBSID'] = srms.OBSID
        return results

    def upload_tokens(self,tokens):
        pass

    def upload_attachments(self,attachment):
        pass

    def modify_fields(self,field_dict):
        for k in field_dict.keys():
            self.tokens.add_keys_to_list(k,field_dict[k])

    def get_staged_srms(self,context):
        ti = context['task_instance']
        srm_xcom = ti.xcom_pull(task_ids=self.staging_task)['srmfile']
        logging.info("Srmfile is "+srm_xcom)
        if srm_xcom==None:
            raise RuntimeError("Could not get the srm list from the "+str(self.staging_task) +" task")
        self.srmlist = srmlist()
        for link in open(srm_xcom,'rb').readlines():
            self.srmlist.append(link)
        return self.srmlist

    def success(self):
        self.status = State.SUCCESS
        logging.info("Successfully uploaded " +
                    str(self.progress['Percent done']) + " % of the tokens.")
    
    def on_kill(self):
        logging.warn('Sending SIGTERM signal to staging group')
        self.state = State.SHUTDOWN
        os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)


class TokenUploader(BaseOperator):
    """
    Using a Token template input, this class creates the tokens for a LOFAR job
    The tokens are a set of documents that contain the metadata for each processing
    job as well as the job's progress, step completion times, and etc. 

    :param sbx_task: The name of the sandbox task which passes the sbx_loc to the tokens
    :type sbx_task: string
    :param srms: a list of the srms that need to be staged
    :type srms: list
    :param stageID: In case staging was already done
    :type stageID: string
    :type output_encoding: output encoding of bash command
    """ 
    template_fields = ()
    template_ext = ()
    ui_color = '#f0ede4'
            
    @apply_defaults
    def __init__(
            self,
            token_task, 
            upload_file=None, 
            parset_task=None,
            pc_database=None,
            output_encoding='utf-8',
            *args, **kwargs):
        
        super(TokenUploader, self).__init__(*args, **kwargs)
        self.pc_database = pc_database
        self.token_task = token_task
        self.parset_task = parset_task
        self.output_encoding = output_encoding
        self.upload_file = upload_file
        self.state = State.QUEUED

    def execute(self, context):
        """
        Execute the bash command in a temporary directory
        which will be cleaned afterwards
        """
        if self.upload_file == None and self.parset_task == None: 
            raise(Exception("No Parset task nor upload file specified!"))
        pc=get_picas_credentials.picas_cred()
        if self.pc_database:
            pc.database = self.pc_database
        tok_dict=context['task_instance'].xcom_pull(task_ids=self.token_task)
        token_id=tok_dict['token_type']
        view=tok_dict['view']
        filename=self.find_upload_file(context)
        self.upload(token_id, view, filename )

    def find_upload_file(self, context):
        """ Checks whether the file exists (if fed a filename as a parameter)
        Otherwise it looks for the xcom of the parset task and takes the 
         key of the dictionary that is named after self.upload_file.split('/')[-1]
         (I.E. the filename) of that xcom. checks if it exists and returns it

        """
        if not self.parset_task and os.path.exists(self.upload_file): 
            return self.upload_file # no parset_task, just get file parameter
        parset_xcom = context['task_instance'].xcom_pull(task_ids=self.parset_task)
        parset_filename=self.upload_file.split('/')[-1]
        parset_file_loc = parset_xcom[parset_filename] 
        if os.path.exists(parset_file_loc): 
            return parset_file_loc
        raise(Exception("Cannot find the parset file"))

    def upload(self, token_id, view, file_name):
        pc=get_picas_credentials.picas_cred()
        if self.pc_database:                                                                                                                                                                                                                                                              
            pc.database = self.pc_database
        th=Token.Token_Handler(t_type=token_id,
                    uname=pc.user,pwd=pc.password,dbn=pc.database)
        self.tokens=th.list_tokens_from_view(view)
        for token in self.tokens:
            th.add_attachment(token.id,open(file_name,'rb'),self.upload_file.split('/')[-1])
        

    def success(self):
        self.status=State.SUCCESS
        logging.info("Successfully uploaded " +
                    str(self.progress['Percent done']) + " % of the tokens.")

    def on_kill(self):
        logging.warn('Sending SIGTERM signal to staging group')
        self.state=State.SHUTDOWN
        os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)



class ModifyTokenStatus(BaseOperator):
    """
    Using a Token template input, this class creates the tokens for a LOFAR job
    The tokens are a set of documents that contain the metadata for each processing
    job as well as the job's progress, step completion times, and etc. 

    :param sbx_task: The name of the sandbox task which passes the sbx_loc to the tokens
    :type sbx_task: string
    :param srms: a list of the srms that need to be staged
    :type srms: list
    :param stageID: In case staging was already done
    :type stageID: string
    :type output_encoding: output encoding of bash command
    """ 
    template_fields = ()
    template_ext = () 
    ui_color = '#f0ede4'
             
    @apply_defaults
    def __init__(
            self,
            token_task,
            pc_database=None,
            modification={'reset':'todo'}, #Dictionary of list of modifications
            output_encoding='utf-8',
            *args, **kwargs):
            
        super(ModifyTokenStatus, self).__init__(*args, **kwargs)
        self.pc_database = pc_database
        self.token_task=token_task
        self.output_encoding = output_encoding
        self.modification=modification
        self.state=State.QUEUED
        
    def execute(self, context):
        """
        Execute the bash command in a temporary directory
        which will be cleaned afterwards
        """
        pc=get_picas_credentials.picas_cred()
        if self.pc_database:
            pc.database = self.pc_database
        tok_dict=context['task_instance'].xcom_pull(task_ids=self.token_task)
        token_id=tok_dict['token_type']
        th=Token.Token_Handler(t_type=token_id,
                    uname=pc.user,pwd=pc.password,dbn=pc.database)        
        for operation, view in self.modification.iteritems():
            if operation=='reset':
                th.reset_tokens(view)
            if operation=='delete':
                th.delete_tokens(view)
            if operation=='set_to_status': # {'set_to_status':{"view":'view_name',"status":'status'}}
                th.set_view_to_status(view['view'],view['status'])

    def success(self):
        self.status=State.SUCCESS
        logging.info("Successfully uploaded " +
                    str(self.progress['Percent done']) + " % of the tokens.")
                    
    def on_kill(self):
        logging.warn('Sending SIGTERM signal to staging group')
        self.state=State.SHUTDOWN
        os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)
        
class ModifyTokenField(BaseOperator):
    """
    Using a Token template input, this class creates the tokens for a LOFAR job
    The tokens are a set of documents that contain the metadata for each processing
    job as well as the job's progress, step completion times, and etc. 

    :param sbx_task: The name of the sandbox task which passes the sbx_loc to the tokens
    :type sbx_task: string
    :param srms: a list of the srms that need to be staged
    :type srms: list
    :param stageID: In case staging was already done
    :type stageID: string
    :type output_encoding: output encoding of bash command
    """
    template_fields = ()
    template_ext = () 
    ui_color = '#f0ede4'

    @apply_defaults 
    def __init__(
            self,
            token_task,
            pc_database=None,
            keyval=['CAL_OBSID','L123456'], #Dictionary of list of modifications
            output_encoding='utf-8',
            *args, **kwargs):
            
        super(ModifyTokenStatus, self).__init__(*args, **kwargs)
        self.pc_database = pc_database
        self.token_task=token_task
        self.output_encoding = output_encoding
        self.modification=modification
        self.state=State.QUEUED
        
    def execute(self, context):
        """
        Execute the bash command in a temporary directory
        which will be cleaned afterwards
        """
        pc=get_picas_credentials.picas_cred()
        tok_dict=context['task_instance'].xcom_pull(task_ids=self.token_task)
        token_id=tok_dict['token_type']
        pc=get_picas_credentials.picas_cred()
        th=Token.Token_Handler(t_type=token_id,
                    uname=pc.user,pwd=pc.password,dbn=pc.database)
        ts=Token.Tokenset(th)
        ts.add_keys_to_list(key=keyval[0],val=keyval[1])

    def success(self):
        self.status=State.SUCCESS
        logging.info("Successfully modified the tokens.")
                    
    def on_kill(self):
        logging.warn('Sending SIGTERM signal to staging group')
        self.state=State.SHUTDOWN
        os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)


class SrmlistFromTokenView(BaseOperator):
    """
    A Token View returns a list of keys/values of (job description) documents
    that match the condition of the view. Using the data returned, we can create
    a srmlist of files stored by these tokens. This will srmlist can be passed
    to a future task to create tokens in chunks as required by the user

    """
    def __init__(
            self,
            token_task,
            view=['test'], 
            pattern='%s_%s', #this will make a srmlist of key_val
            output_encoding='utf-8',
            *args, **kwargs):
        
        super(ModifyTokenStatus, self).__init__(*args, **kwargs)
        self.token_task=token_task
        self.view=view
        self.pattern=pattern
        self.output_encoding = output_encoding
        self.state=State.QUEUED
        
    def execute(self, context):  
        tokens=self.get_tokens_from_view()
        f=NamedTemporaryFile(delete=False)
        surl_list=self.make_srmlist_from_token_list(tokens,
                pattern=self.pattern)
        for i in surl_list:
            f.write(i)

        return f.name

    def get_tokens_from_view(self):
        tokens=[]
        tok_dict=context['task_instance'].xcom_pull(task_ids=self.token_task)
        token_id=tok_dict['token_type']
        token_view=tok_dict['view'] #TODO: Decide if use view here or in initializer
        pc=get_picas_credentials.picas_cred()
        th=Token.Token_Handler(t_type=token_id,
                uname=pc.user,pwd=pc.password,dbn=pc.database)
        for t in th.list_tokens_from_view(token_view):
            tokens.append(t)
        return tokens

    def make_srmlist_from_token_list(self,token_list=[],pattern="%s_%s"):
        surl_list=srmlist.srmlist()
        for token in token_list:
            surl_list.append(pattern % (token['key'], token['value']))
        return surl_list


class TokenArchiver(BaseOperator):
    """
    Archives all data from a certain run, given token_type and output directory
                                                                                                         
    :param token_type: The task that returns the token_type
    :type token_type: string
    :type output_encoding: output encoding of bash command
    """ 
    template_fields = ()
    template_ext = ()
    ui_color = '#c563ff'
            
    @apply_defaults
    def __init__(
            self,
            token_type_task,
            base_archive_location = 'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/distrib/SKSP',
            output_encoding = 'utf-8',
            *args, **kwargs):
        
        super(TokenArchiver, self).__init__(*args, **kwargs)
        self.output_encoding = output_encoding
        self.base_archive_location = base_archive_location
        self.t_task = token_type_task
        self.state = State.QUEUED
        
    def execute(self, context):
        self.tmpdir=mkdtemp() #Makes temporary directory
        self.oldpwd=os.getcwd()
        tok_dict=context['task_instance'].xcom_pull(task_ids=self.t_task)
        token_type = tok_dict['token_type']              
        self.OBSID = tok_dict['OBSID']
        pc=get_picas_credentials.picas_cred()
        self.th=Token.Token_Handler(t_type=token_type,
            uname=pc.user,pwd=pc.password,dbn=pc.database)
        archive_file=self.archive_tokens()
        self.upload_archive_to_storage(archive_file)
        self.th.purge_tokens()

    def upload_archive_to_storage(self,archive):
        dest_dir = self.base_archive_location+"/"+str(self.OBSID)
        upload = Popen(['globus-url-copy', archive, dest_dir+"/"+"token_archive.tar.gz"], stdout=PIPE, stderr=PIPE)
        execute = upload.communicate()
        if execute[1] != '':
            logging.warn("Upload error?")
            logging.warn(execute[1])

    def archive_tokens(self):
        os.chdir(self.tmpdir)
        self.th.archive_tokens(delete_on_save=True)
        pc=get_picas_credentials.picas_cred()
        self.tarfile = "tokens_"+pc.user+"_"+pc.database+"_"+self.th.t_type+".tar.gz"
        with tarfile.open(self.tmpdir+"/"+self.tarfile, "w:gz") as tar:
            tar.add(self.tmpdir)
        os.chdir(self.oldpwd)
        return self.tmpdir+"/"+self.tarfile

    def remove_tmpdir(self):
        shutil.rmtree(self.tmpdir)
        




#!/usr/bin/env python

#from distutils.core import setup
from setuptools import setup
import os

setup(name='AGLOW',
      packages=['AGLOW','AGLOW/airflow','AGLOW/airflow/dags','AGLOW/airflow/operators','AGLOW/airflow/sensors','AGLOW/airflow/utils'], #'GRID_LRT', 'GRID_LRT/Staging', 'GRID_LRT/Application', 'GRID_LRT/couchdb'],
      version='v0.0.9.1',
      setup_requires=[
        'apache-airflow',
        'grid-lrt'
          ],
      tests_require=[
        'pytest', 
          ],
      include_package_data=True,
      data_files = [ 
          ],
      description='AGLOW: Automated Grid-enabled LOFAR Workflows',
      long_description="Distributing, Automating and Accelerating LOFAR data processing using an industry standard workflow orchestration software on a High Throughput cluster at the Dutch Grid location.",
      author='Alexandar Mechev',
      author_email='LOFAR@apmechev.com',
      url='https://www.github.com/apmechev/AGLOW/',
      download_url = 'https://github.com/apmechev/AGLOW/archive/v0.0.9x.tar.gz',
      keywords = ['surfsara', 'distributed-computing', 'LOFAR', 'workflow','airflow'],
      classifiers = ['Development Status :: 2 - Pre-Alpha',
      'Intended Audience :: Developers',
      'Intended Audience :: Science/Research',
      "License :: OSI Approved :: GNU General Public License v3 (GPLv3)" ,
      "Natural Language :: English",
      "Topic :: Scientific/Engineering :: Astronomy",
      "Topic :: System :: Distributed Computing",
#      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 2.6',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.2',
      'Programming Language :: Python :: 3.3',
      'Programming Language :: Python :: 3.4'] 
     )

